/* OS/2 REXX: create a list of all EA .TYPE found on local disks, */
/* use CHECKTYP without argument to get a .TYPE-report, see below */
/* usage: CHECKTYP [REDUCE|REPAIR|REMOVE|type]                    */

/* Observations:                                                  */
/* - .TYPE should be MVMT (0xFFDF), but often it's ASCII (0xFFFD) */
/* - an early test showed 965 MVMT vs. 646 ASCII, in a later test */
/*   917 of 940 MVMT contained a single ASCII item, the remaining */
/*   23 MVMT contained two ASCII items                            */
/* - NetScape 2.02 uses ASCII for its MIME types, including weird */
/*   .TYPEs like "application/x-unknown-content-type" (34 char.s) */
/* - obviously (ckecked with WPS) it's possible to reduce MVMT to */
/*   ASCII if it contains only one item =>  call this script with */
/*   argument REDUCE to replace MVMT by ASCII where possible      */
/* - argument REDUCE only replaces MVMT with codepage 0 by ASCII, */
/*   on my system all .TYPE MVMT EAs used the default codepage 0  */
/* - argument REDUCE corrected numerous length errors implicitly, */
/*   to correct more errors explicitly use argument REPAIR (this  */
/*   also removes invalid .TYPE EAs, e.g. neither ASCII nor MVMT) */
/* - REDUCE and REPAIR implicitly corrected cases, where MVMT EA  */
/*   contained a single additional NUL byte, shown as "error x00" */
/* - note that REPAIR cannot yet fix all possible MVMT problems,  */
/*   but on my system it was enough to REPAIR all MVMT problems,  */
/*   see also the restriction for REDUCE (codepage 0 and 1 item)  */
/* - argument REMOVE deletes ASCII .TYPEs of the form ?/x-? like  */
/*   application/x-gzip or image/x-icon (unregistered MIME type)  */
/* - please check <WP_NOWHERE> and NETSCAPE.INI files manually to */
/*   get rid of Netscape's associations with REMOVEd ?/x-? .TYPEs */
/* - to find more than 1 file using any given .TYPE specify it as */
/*   argument (e.g. PLAIN TEXT, case insensitive)                 */
/* - you cannot match a void .TYPE (i.e. ASCII length 0), this is */
/*   done automatically if no argument is specified - you'll then */
/*   find counters like 1 MT+: or 2 CII: in the report instead of */
/*   the usual tags MVMT+: resp. ASCII: (see below report format) */
/* - converting ASCII to MVMT is simple, x2c( FD FF 00 00 01 00 ) */
/*   added before any valid ASCII EA - sorry, not yet implemented */

/* Report format:                                                 */
/* - report lines in the form "MVMT n item(s)" plus COUNTER show  */
/*   a counter of MVMT .TYPE EAs with n=1 etc. item(s), otherwise */
/*   report lines are either TAG TYPE FILE or TAG TYPE COUNTER if */
/*   TAG TYPE matches more than one file.  The path of a FILE is  */
/*   truncated if necessary (first 6 char.s, ..., last 25 char.s) */
/* - the TYPE in a report line is the ASCII value truncated to 37 */
/*   char.s starting in column 8 (control char.s replaced by ?) - */
/* - a TYPE starting in column 7 with x is a hex. translation for */
/*   unknown binary values (e.g. not ASCII or missing MVMT count) */
/* - 123456789, possible TAGs (normally columns 1..7):            */
/*   ASCII:     valid ASCII .TYPE, i.e. not part of a MVMT EA     */
/*   ASCII?     dito, but invalid length (maybe use REPAIR)       */
/*   ASCII=     dito, REDUCEd (was MVMT) or REPAIRed (length)     */
/*   MVMT+:     valid ASCII in MVMT .TYPE EA                      */
/*   MVMT+?     dito, but invalid length (maybe try REPAIR)       */
/*   x???? x    neither ASCII nor MVMT, REPAIR would erase EA     */
/*   xFFFE x    e.g. xFFFE (EAT BINARY) is not okay for .TYPE     */
/*   m???? x    non-ASCII in MVMT, REPAIR would erase MVMT EA     */
/*   mFFFE x    e.g. mFFFE (hex. value follows x in column 7)     */
/*   MVMT? x    MVMT too short, upto 3 hex. bytes following x     */
/*   locked     REMOVE, REPAIR, or REDUCE failed (OS/2 rc 32)     */
/*   erased     REMOVEd, old TYPE not shown, only the COUNTER     */
/*   MVMT n     summary of MVMT EAs with n item(s), examples:     */
/*   MVMT 1 item(s) .............. 7        (can be REDUCEd)      */
/*   MVMT 1 item(s) in CP 861 .... 3        (codepage not 0)      */
/*   MVMT 2 item(s) ............. 14        (=> irreducible)      */
/*   missed n   found only k of n + k MVMT items, REPAIR does not */
/*              yet handle this or other "irreducible" cases      */
/*   error x    remaining hex. data after analyzing all items     */
/*   error x00  e.g. trailing NUL after otherwise valid MVMT, in  */
/*              "reducible" cases REDUCE would get rid of the NUL */
/*   n ????     n-th void (length 0) ASCII value (overlaying tag) */
/*   1 CII:     e.g. 1st void value overlaying tag ASCII:         */
/*   2 MT+:     e.g. 2nd void value overlaying tag MVMT+:         */
/*   (???):     n-th (???) value of .TYPE matched by argument     */
/*   (001):     e.g. 1st found ASCII or MVMT .TYPE value          */
/*   (002)?     e.g. 2nd found value (? in column 6: bad length)  */
/* - to sort a report by type values simply use columns 8 upto 44 */

/* Later fixes and additional observations:                       */
/* Apparently MVMT .TYPE EAs with zero items are allowed, but the */
/* old script version treated this as "irreducible".  Now REDUCE  */
/* and REPAIR delete empty MVMT .TYPE EAs for default codepage 0. */

/* An application created MVMT .TYPE EAs with 2 items specifying  */
/* only one ASCII value (reported as "missed 1").  To display the */
/* corresponding files use "missed 1" as argument.  Don't expect  */
/* this "undocumented feature" to work in future script versions. */

   signal on  novalue  name TRAP    ;  signal on  syntax name TRAP
   signal on  failure  name TRAP    ;  signal on  halt   name TRAP
   call UTIL 'SysGetEA'             ;  call UTIL 'SysFileTree'
   call UTIL 'SysPutEA'             ;  ALL. = 0    ;  arg XARG
   ISVIO = ( PROC() < 3 )           ;  EXPO = 'ISVIO ALL. EAS. XARG'

   parse source . . TMP             ;  N = lastpos( '\', TMP )
   TMP = substr( TMP, N + 1 )       ;  N = lastpos( '.', TMP )
   TMP = left( TMP, N ) || 'tmp'    ;  N = XENV( 'TMP' )
   if N = '' then N = '.'           ;  TMP = N || '\' || TMP
   call XDEL TMP                    ;  call CTOZ

   if ISVIO then say d2c( 13 ) || left( 'writing' TMP || '...', 79 )

   do N = 1 to ALL.0
      P = ALL.N
      if length( EAS.P ) > 34
         then  P = left( EAS.P, 6 ) || '...' || right( EAS.P, 25 )
         else  P =                              right( EAS.P, 34 )
      call lineout TMP, left( ALL.N || ' ', 44, . ) P
   end N

   exit 0

CTOZ: procedure expose (EXPO)    /* call SUBS on all local drives */
   call UTIL 'SysDriveMap'          ;  call UTIL 'SysFileSystemType'
   D = SysDriveMap( /**/, 'Local' )

   do N = 1 to words( D )        /* no EA access on CDFS (why ?)  */
      if SysFileSystemType( word( D, N )) = 'CDFS' then  iterate N
      if SUBS(              word( D, N ))          then  return 1
   end N
   return 0

SUBS: procedure expose (EXPO)
   FAIL = 0 ;  SUBS = arg( 1 ) || '\*'
   if sign( SysFileTree( SUBS, 'SUB', 'DO' )) then drop SUB.
   do N = 1 to SUB.0 ;  FAIL = FAIL | SUBS( SUB.N )   ;  end N
   do N = 1 to SUB.0 ;  FAIL = FAIL | DOIT( SUB.N )   ;  end N

   if sign( SysFileTree( SUBS, 'SUB', 'FO' )) then drop SUB.
   do N = 1 to SUB.0 ;  FAIL = FAIL | DOIT( SUB.N )   ;  end N

   return FAIL                   /* 1: errors are simply ignored  */

DOIT: procedure expose (EXPO)    /* 1: error for e.g. EA DATA. SF */
   if ISVIO then call charout /**/, x2c( 0D ) || left( arg(1), 79 )
   ASCII = reverse( x2c( 'FFFD' ))
   MULTI = reverse( x2c( 'FFDF' ))

   THIS = SysGetEA( arg( 1 ), '.TYPE', 'VAL' )
   select
      when THIS = 32             then return 0
      when THIS <> 0             then exit TRAP( 'error' THIS )
      when VAL  = ''             then return 0
      when abbrev( VAL, ASCII )  then do
         N = c2d( reverse( substr( VAL, 3, 2 )))
         VAL = substr( VAL, 5 )           /* ASCII: if length ok. */
         if N = length( VAL ) then  N = ':'  ;  else N = '?'
         if N = ':' & XARG = 'REMOVE' then do
            parse var VAL X '/x-' Y       /* if unregistered MIME */
            if X <> VAL & Y <> '' then return SEAT( arg( 1 ))
         end
         if N = '?' & XARG = 'REPAIR'     /* ASCII? if length bad */
            then return SEAT( arg( 1 ), strip( VAL, 'T', d2c( 0 )))
            else return EADD( arg( 1 ), 'ASCII' || N VAL )
      end
      when abbrev( VAL, MULTI )  then do
         if length( VAL ) < 6 then do     /* MVMT? x345 too short */
            THIS = 'MVMT? x' || c2x( substr( VAL, 3 ))
            if XARG = 'REPAIR'
               then return SEAT( arg( 1 ))
               else return EADD( arg( 1 ), THIS )
         end

         CODE = c2d( reverse( substr( VAL, 3, 2 )))
         CNT  = c2d( reverse( substr( VAL, 5, 2 )))
         VAL  = substr( VAL, 7 )
         if CODE = 0                      /* MVMT summary (items) */
            then call EADD arg( 1 ), 'MVMT' CNT 'item(s)'
            else call EADD arg( 1 ), 'MVMT' CNT 'item(s) in CP' CODE

         do N = 1 to CNT until VAL == ''  /* parse single MVMT EA */
            parse var VAL EAT =3 LEN =5 VAL
            if EAT <> ASCII then do       /* mXXXX MVMT not ASCII */
               THIS = 'm' || c2x( reverse( EAT ))
               THIS = THIS 'x' || c2x( LEN || VAL )
               if XARG = 'REPAIR'
                  then return SEAT( arg( 1 ))
                  else return EADD( arg( 1 ), THIS )
            end

            LEN = c2d( reverse( LEN ))    /* MVMT+: if next ASCII */
            BAD = ( LEN > length( VAL ))  /* MVMT+? if bad length */
            if BAD   then  THIS = '?' VAL
                     else  THIS = ':' left( VAL, LEN )
            if CNT = 1 & CODE = 0 then do /* can REDUCE or REPAIR */
               BAD = BAD & ( XARG = 'REPAIR' )
               if    BAD | ( XARG = 'REDUCE' ) then do
                  return SEAT( arg( 1 ), strip( VAL, 'T', d2c( 0 )))
               end                        /* remove trailing zero */
            end                           /* and replace ASCII EA */

            call EADD arg( 1 ), 'MVMT+' || THIS
            VAL = substr( VAL, LEN + 1 )
         end N

         if CNT <= N & VAL == '' then  select
            when CNT = N         then  return 0
            when CNT > 0         then  return TRAP( 'script error' )
            when CODE <> 0       then  return 0
            when XARG = 'REDUCE' then  return SEAT( arg( 1 ))
            when XARG = 'REPAIR' then  return SEAT( arg( 1 ))
            otherwise                  return 0
         end
         if CNT > N
            then return EADD( arg( 1 ), 'missed' CNT - N )
            else return EADD( arg( 1 ), 'error x' || c2x( VAL ))
      end
      otherwise                  /* show unexpected EAT as binary */
         THIS = 'x' || c2x( reverse( left( VAL, 2 )))
         THIS = THIS 'x' || c2x( substr( VAL, 3 ))
         if XARG = 'REPAIR'
            then return SEAT( arg( 1 ))
            else return EADD( arg( 1 ), THIS )
   end

SEAT: procedure expose (EXPO)    /* put ASCII or remove .TYPE EA: */
   VAL = ''
   if arg() = 2 then do
      VAL = VAL || reverse( x2c( 'FFFD' ))
      VAL = VAL || reverse( x2c( d2x( length( arg( 2 )), 4 )))
      VAL = VAL || arg( 2 )
   end

   R = SysPutEA( arg( 1 ), '.TYPE', VAL )
   select
      when  R = 32   then  return EADD( arg( 1 ), 'locked' arg( 2 ))
      when  R <> 0   then  return TRAP( 'error' R 'put EA' arg( 1 ))
      when  VAL = '' then  return EADD( arg( 1 ), 'erased' arg( 2 ))
      otherwise            return EADD( arg( 1 ), 'ASCII=' arg( 2 ))
   end                           /* TRAP returns to create report */

EADD: procedure expose (EXPO)
   THIS = arg( 2 )
   THIS = translate( THIS, /**/, xrange( d2c( 0 ), d2c( 31 )), '?' )

   /* -- tricky part to mark .TYPEs wanted by user differently -- */
   X = sign( wordpos( XARG, 'REMOVE REPAIR REDUCE' )) | XARG = ''
   do N = 1 to 1 - X             /* i.e. try match once or never: */
      X = abbrev( reverse( translate( arg( 2 ))), reverse( XARG ))
      if X = 0 | length( THIS ) - length( XARG ) > 7 then leave N
      ALL.. = ALL.. + 1          /* gotcha, make THIS id. unique: */
      X = '(' || right( ALL.., 3, 0 ) || ')' || substr( X, 6 )
      THIS = X || substr( THIS, 1 + length( X ))
   end N                         /* unique id. noted as 1st below */
   if XARG = '' & length( THIS ) = 7 then do
      ALL.. = ALL.. + 1          /* void EA, make THIS id. unique */
      THIS = overlay( ALL.. || ' ', THIS )
   end
   /* ----------------------------------------------------------- */

   if symbol( 'EAS.THIS' ) = 'VAR' then do
      if datatype( EAS.THIS, 'w' )
         then EAS.THIS = EAS.THIS + 1
         else EAS.THIS = 2       /* 2nd: path replaced by counter */
   end
   else do                       /* 1st: note value in next ALL.N */
      N = ALL.0 + 1  ;  EAS.THIS = arg( 1 )
      ALL.0 = N      ;  ALL.N    = THIS
   end

   return 0

/* see <URL:http://purl.net/xyzzy/rexxtrap.htm>, (c) F. Ellermann */

XDEL: procedure                  /* semi-portable SysFileDelete() */
   parse source KEY . .
   if KEY = 'OS/2' then do
      call UTIL 'SysFileDelete'
      return SysFileDelete( arg( 1 )) <> 0   /* 0: okay, 1: error */
   end
   parse version KEY . .
   if KEY = 'REXXSAA'
      then return RxDelete( arg( 1 )) <> 0   /* 0: okay, 1: error */
      else return DosDel(   arg( 1 )) == 0   /* 0: okay, 1: error */

XENV: procedure                  /* DOS REXX portable environment */
   parse version ENV . .
   if ENV = 'REXXSAA' then do
      parse source ENV . .       /* OS/2 REXXSAA:  os2environment */
      if ENV = 'OS/2'   then ENV = 'OS2ENVIRONMENT'
                        else ENV = ENV || 'ENVIRONMENT'
   end                           /* DOS  REXXSAA:  DOSENVIRONMENT */
   else  ENV = 'ENVIRONMENT'     /* REXX/Personal:    environment */
   select
      when arg() = 1 then  return value( arg( 1 ),/* get */, ENV )
      when arg() = 2 then  return value( arg( 1 ), arg( 2 ), ENV )
      otherwise            return abs( /* force REXX error 40 */ )
   end

PROC: procedure                  /* avoid "unknown function" TRAP */
   parse source OS .             /* for REXXSAA portability abuse */
   if OS <> 'OS/2' then return 1 /* the now obsolete 1: real mode */

   OS = 'ProcessType'            /* assume Sys... = RxProcessType */
   if RxFuncQuery( 'Sys' || OS ) = 0 then return SysProcessType()
   if RxFuncAdd(   'Sys' || OS, 'RxUtils', 'Rx' || OS ) = 0 then do
      signal on syntax name PROC.TRAP  ;  return SysProcessType()
   end                           /* tries RxUtils only once, else */
PROC.TRAP:                       /* force RexxUtil SysProcessType */
   call  RxFuncDrop 'SysProcessType'   ;  signal on syntax name TRAP
   call        UTIL 'SysProcessType'   ;  return SysProcessType()

UTIL: procedure                  /* load necessary RexxUtil entry */
   if RxFuncQuery(  arg( 1 )) then
      if RxFuncAdd( arg( 1 ), 'RexxUtil', arg( 1 )) then
         exit TRAP( "can't add RexxUtil"  arg( 1 ))
   return 0

TRAP:                            /* select REXX exception handler */
   call trace 'O' ;  trace N           /* don't trace interactive */
   parse source TRAP                   /* source on separate line */
   TRAP = x2c( 0D ) || right( '+++', 10 ) TRAP || x2c( 0D0A )
   TRAP = TRAP || right( '+++', 10 )   /* = standard trace prefix */
   TRAP = TRAP strip( condition( 'c' ) 'trap:' condition( 'd' ))
   select
      when wordpos( condition( 'c' ), 'ERROR FAILURE' ) > 0 then do
         if condition( 'd' ) > ''      /* need an additional line */
            then TRAP = TRAP || x2c( 0D0A ) || right( '+++', 10 )
         TRAP = TRAP '(RC' rc || ')'   /* any system error codes  */
         if condition( 'c' ) = 'FAILURE' then rc = -3
      end
      when wordpos( condition( 'c' ), 'HALT SYNTAX'   ) > 0 then do
         if condition( 'c' ) = 'HALT' then rc = 4
         if condition( 'd' ) > '' & condition( 'd' ) <> rc then do
            if condition( 'd' ) <> errortext( rc ) then do
               TRAP = TRAP || x2c( 0D0A ) || right( '+++', 10 )
               TRAP = TRAP errortext( rc )
            end                        /* future condition( 'd' ) */
         end                           /* may use errortext( rc ) */
         else  TRAP = TRAP errortext( rc )
         rc = -rc                      /* rc < 0: REXX error code */
      end
      when condition( 'c' ) = 'NOVALUE'  then rc = -2 /* dubious  */
      when condition( 'c' ) = 'NOTREADY' then rc = -1 /* dubious  */
      otherwise                        /* force non-zero whole rc */
         if datatype( value( 'RC' ), 'W' ) = 0 then rc = 1
         if rc = 0                             then rc = 1
         if condition() = '' then TRAP = TRAP arg( 1 )
   end                                 /* direct: TRAP( message ) */

   TRAP = TRAP || x2c( 0D0A ) || format( sigl, 6 )
   signal on syntax name TRAP.SIGL     /* throw syntax error 3... */
   if 0 < sigl & sigl <= sourceline()  /* if no handle for source */
      then TRAP = TRAP '*-*' strip( sourceline( sigl ))
      else TRAP = TRAP '+++ (source line unavailable)'
TRAP.SIGL:                             /* ...catch syntax error 3 */
   if abbrev( right( TRAP, 2 + 6 ), x2c( 0D0A )) then do
      TRAP = TRAP '+++ (source line unreadable)'   ;  rc = -rc
   end
   select
      when 0 then do                   /* in pipes STDERR: output */
         parse version TRAP.REXX . .   /* REXX/Personal: \dev\con */
         signal on syntax name TRAP.FAIL
         if TRAP.REXX = 'REXXSAA'      /* fails if no more handle */
            then call lineout 'STDERR'  , TRAP
            else call lineout '\dev\con', TRAP
      end
      when 0 then do                   /* OS/2 PM: RxMessageBox() */
         signal on syntax name TRAP.FAIL
         call RxMessageBox ,           /* fails if not in PMREXX  */
            translate( TRAP, ' ', x2c( 0D )), , 'CANCEL', 'WARNING'
      end                              /* replace any CR by blank */
      otherwise   say TRAP ; trace ?L  /* interactive Label trace */
   end

   if condition() = 'SIGNAL' then signal TRAP.EXIT
TRAP.CALL:  return rc                  /* continue after CALL ON  */
TRAP.FAIL:  say TRAP ;  rc = 0 - rc    /* force TRAP error output */
TRAP.EXIT:  exit   rc                  /* exit for any SIGNAL ON  */
